#
#             RPKP Experiment Log Retrieval and Analysis
#
#              by John Walker  --  kelvin@fourmilab.ch
#                      http://www.fourmilab.ch/
#
#               This program is in the public domain.
#
#   This program is designed to be executed by Perl 4.036.

    require "splitfields.pl";
    require "getopts.pl";
    require "combi.pl";

    $Logfile = "/files/server/logs/rpkp/rpkp.log";
    $Raw = 1;                         # Return raw data from run log ?
    $Html = 0;                        # HTML format output ?
    $All = 0;                         # Return all data, or just one subject ?
    $Practice = 0;                    # Include practice mode runs
    $Watched = "";                    # E-mail ID of selected subject
    $Bowdler = 0;                     # Strip user and site ID to preserve privacy ?
    $Dump = 0;                        # Dump scanned fields for debugging
    $Debug = 0;                       # Debug mode output (can compromise user
                                      # privacy--do not allow remote clients to
                                      # set this)

    $subject = 0;                     # No subject numbers assigned yet
    $subjects{''} = '_';              # No E-mail address maps to blank subject
    @nbits = (0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4); # Bits set in nybble

    #   Process command line options

    &Getopts('i:e:abdghpr');

    if (defined($opt_a)) {
        $All = 1;                     # -a   Return records for all runs
    }
    if (defined($opt_b)) {
        $Bowdler = 1;                 # -b   Elide private user information
    }
    if (defined($opt_d)) {
        $Dump = 1;                    # -d   Dump log file records
    }
    if (defined($opt_e)) {
        $Watched = $opt_e;            # -e   Email-address-of-subject
    }
    if (defined($opt_g)) {
        $Debug = 1;                   # -g   Generate debug output
    }
    if (defined($opt_h)) {
        $Html = 1;                    # -h   HTML format output
        $Raw = 0;
    }
    if (defined($opt_i)) {
        $Logfile = $opt_i;            # -i   Logfile
    }
    if (defined($opt_p)) {
        $Practice = 1;                # -p   Include Practice mode runs
    }
    if (defined($opt_r)) {
        $Raw = 1;                     # -r   Return raw format records
        $Html = 0;
    }

    #   If we're Bowdlerising user ID information, and the user has
    #   specified their own E-mail ID, map their own address into
    #   "Me" rather than a "Subject" designation.  This allows a user
    #   to identify their own runs even when retrieving the entire
    #   database.

    if ($Watched ne "") {
        $subjects{$Watched} = "Me";
    }

    if (!$Html) {
        print("Content-type: text/plain\nTitle: RetroPsychoKinesis Experiment Log\n\n");
    }

    open(LOG, "<" . $Logfile) || die "Unable to open $Logfile: $!\n";

    while ($DBitem = <LOG>) {
        chop($DBitem);
        ($Format, $Live, $Experiment, $Mode, $Goal, $Sound,
         $Email, $IPaddress, $Hostname, $Date, $Stream) = &splitfields($DBitem);

        #   If selecting records for a given subject, discard those
        #   for others.

        if (!$All && ($Email ne $Watched)) {
            next;
        }

        #   Unless specifically requested, discard Practice mode runs

        if (!$Practice && ($Mode eq 'Practice')) {
            next;
        }

        if ($Dump) {
            ($zzt, $zzo) = &count_bits($Stream);
            ($sec, $min, $hour, $mday, $mon, $year) = gmtime($Date);
            print <<"EOH"
Format: $Format  Live? $Live  Experiment: $Experiment  Mode: $Mode
         Goal: $Goal  Sound: $Sound  E-mail: $Email
         IP: $IPaddress  Hostname: $Hostname  Date: $year/$mon/$mday $hour:$min UTC
         Bits: $zzt  Ones: $zzo  Data: $Stream
EOH
            ;
        }

        #   If we're Bowdlerising and this is an E-mail address we
        #   haven't encountered previously, assign it a new subject
        #   identifier.

        if ($Bowdler) {
            if (!defined($subjects{$Email})) {
                $subject++;
                $subjects{$Email} = "Subject_$subject";
            }
        }

        if ($Raw) {
            if ($Bowdler) {
                #   Note that we don't have to worry about quoting any fields
                #   in Bowdlerised output since all the fields which might
                #   contain troublemaking characters have been dropped.
                print("$Format,$Live,$Experiment,$Mode,$Goal,$Sound,$subjects{$Email},,,$Date,$Stream\n");
            } else {
                $em = &quote_string($Email);
                $hn = &quote_string($Hostname);
                print("$Format,$Live,$Experiment,$Mode,$Goal,$Sound,$em,$IPaddress,$hn,$Date,$Stream\n");
            }
        } elsif ($Html) {
            if ($Bowdler) {
                $id = $subjects{$Email};
                if ($id =~ m/^Subject_/) {
                    $id = sprintf("Subject %12ld", substr($id, 8));
                }
            } else {
                $id = $Email;
            }
            push(@record, sprintf("%-48s %11lu", $id, $Date) . $DBitem);
        } else {
            print("No output format option specified!\n");
        }
    }

    if ($Html) {
        print <<"EOH"
Content-type: text/html

<html>
<head>
<title>RetroPsychoKinesis Experiment Log Report</title>
</head>
<body>
<center>
<h1><a href="http://www.fourmilab.ch/rpkp/">RetroPsychoKinesis</a> Experiment Log Report </h1>
</center>
<hr>
EOH
        ;
        if ($All) {
            print("<h1>Results by Subject</h1>\n");
            print("The tables below show results from all RPKP experiments\n");
            if ($Practice) {
                print("in both Record and Practice");
            } else {
                print("in Record");
            }
            print(" mode since inception of the project, sorted by\n");
            print("the subject's identity and date and time of the run.\n");
            print("The E-mail or made-up \"handle\" entered by a subject\n");
            print("has been replaced by a Subject number to preserve\n");
            print("the subjects' privacy.  If you've entered your E-mail\n");
            print("address (or handle) in the form requesting this report,\n");
            print("the table containing your own runs will be titled \"Me\".\n");
        } else {
            print("<h1>Your Runs To Date</h1>\n");
            print("The table below shows results from all RPKP experiment runs\n");
            if ($Practice) {
                print("in both Record and Practice");
            } else {
                print("in Record");
            }
            print(" mode you've made to date.\n");
        }

        print <<"EOH"

<center>
EOH
        ;
        $intable = 0;
        $Grand_experiments = 0;
        $Grand_subjects = 0;
        $Grand_tries = 0;
        $Grand_hits = 0;
        foreach $r (sort @record) {
#           print "$r\n";
            ($Format, $Live, $Experiment, $Mode, $Goal, $Sound,
             $Email, $IPaddress, $Hostname, $Date, $Stream) = &splitfields(substr($r, 60));
             $Grand_experiments++;
             if ($Bowdler) {
                $Email = $subjects{$Email};
                $Email =~ s/_/\ /;
             }
            ($zzt, $zzo) = &count_bits($Stream);
            if (substr($Goal, 1, 1) eq ':') {
                $Goal = substr($Goal, 0, 1);
            }
            if ($Goal == 0) {
                $zzo = $zzt - $zzo;
            }
            $Grand_tries += $zzt;
            $Grand_hits += $zzo;
            $sd = (($zzt / 2) - $zzo) / &standard_deviation($zzt);
            if ($sd < 0) {
                $sd = -$sd;
            }
            $sd = sprintf("%.2f", $sd);
            $prob = &chance($zzo, $zzt, 0.5);
            if ($zzo < ($zzt / 2)) {
                $sd = "<font color=\"#FF0000\">" . $sd . "</font>";
            }
            ($sec, $min, $hour, $mday, $mon, $year) = gmtime($Date);
            $td = sprintf("%d/%d/%d %2d:%02d", $year + 1900, $mon + 1, $mday,
                            $hour, $min);
            $so = ($Sound eq '-s') ? 'Y' : '&nbsp;';
            if (!$intable || ($lastuser ne $Email)) {
                if ($intable) {
                    &end_table();
                }
                $Grand_subjects++;
                $intable = 1;
                $lastuser = $Email;
                $Ttries = 0;
                $Thits = 0;
                if ($Email eq ' ') {
                    $Email = 'User ID unspecified';
                }
                print <<"EOH"
<h3>$Email</h3>
<table border cellpadding=4>
<tr>
<th>Date / Time
<th>Experiment
<th>Goal
<th>Mode
<th>Sound?
<th>Hits
<th>Sigma
<th>Chance
</tr>
EOH
                ;
            }
            $Ttries += $zzt;
            $Thits += $zzo;
            print <<"EOH"
<tr>
<td align=center>$td
<td align=center>$Experiment
<td align=center>$Goal
<td align=center>$Mode
<td align=center>$so
<td align=center>$zzo/$zzt
<td align=center>$sd
<td align=center>$prob
</tr>
EOH
            ;

        }
        if ($intable) {
            &end_table();
        }
        print("</center>\n");

        print <<"EOH"
<h3>Columns in the Table</h2>
<dl>
<dt><b>Date/Time</b>
<dd>Time at which the experiment was run
in Universal (Greenwich Mean) Time.
<p>
<dt><b>Experiment</b>
<dd>Which
of the visual feedback programs was used.
<p>
<dt><b>Goal</b>
<dd>Whether
the subject was attempting to increase the number of one bits or zero bits
(selected as an option on the visual feedback program).
<p>
<dt><b>Mode</b>
<dd>Whether the run was in Record or Practice mode (Practice
mode runs are not normally shown, but can be viewed by checking a
box requesting them in the report request form).
<p>
<dt><b>Sound</b>
<dd>\"Y\" if audible feedback was requested.
<p>
<dt><b>Hits</b>
<dd>How many occurrences of the bit given by
<b>Goal</b> appeared in the bit stream from the random number
generator; the most probable chance results cluster around half the
length of the sequence: roughly equal numbers of zero and one bits.
<p>
<dt><b>Sigma</b>
<dd>In standard deviations, the extent to which
the number of hits diverges from the expectation due to chance alone.
The larger the <b>Sigma</b>, the more unlikely the result is to occur
purely by chance.  Sigma is shown in black if more hits than misses
occurred, and red otherwise.  The standard deviation is identical for
the number of hits and misses.
<p>
<dt><b>Chance</b>
<dd>Assuming the bit stream is due only to chance, how many runs you would
have to make, on average, for the results of the actual run to recur.
If you managed, for example, to make 588 hits in a run of 1024
bits, you could only expect, by pure chance, to score that high again
after making 877647 more runs.  Consistently high Sigma and Chance
figures for a given subject support the hypothesis that the subject
is influencing the results of the experiment.
</dl>
EOH
        ;
    }

    #   If Bowdlerising and debugging, show the mapping between subject
    #   names and E-mail addresses.

    if ($Bowdler && $Debug) {
        if ($Html) {
            print("<h3>Subject to E-mail mapping</h3>\n<pre>\n");
        } else {
            print("\n");
        }
        foreach $subj (sort keys(%subjects)) {
            $sname = $subjects{$subj};
            if ($sname ne '' && $subj ne '') {
                printf("%-20s %s\n", $sname, $subj);
            }
        }
        if ($Html) {
            print("</pre>\n");
        }
    }

    if ($Html) {
        $gsd = (($Grand_tries / 2) - $Grand_hits) / &standard_deviation($Grand_tries);
        if ($gsd < 0) {
            $gsd = -$gsd;
        }
        $gsd = sprintf("%.2f", $gsd);
        if ($Grand_hits < ($Grand_tries / 2)) {
            $gsd = "<font color=\"#FF0000\">" . $gsd . "</font>";
        }

        if ($All) {
            print <<"EOH"
<h1>Overall Summary</h1>
<table> 
<tr><td>Total experiments:<td align=right>$Grand_experiments
<tr><td>Number of subjects:<td align=right>$Grand_subjects
<tr><td>Total tries:<td align=right>$Grand_tries
<tr><td>Total hits:<td align=right>$Grand_hits
<tr><td>Overall sigma:<td align=right>$gsd
</table>
EOH
            ;
        }

        print <<"EOH"
<h3><a href="/rpkp/experiments/analysis/">New Log Report</a></h3>

<h3><a href="/rpkp/experiments/">RetroPsychoKinesis Experiments Online</a></h3>

<h3><a href="/rpkp/"><img src="/rpkp/images/home.gif" width=97 height=64 align=middle alt=""></a> <a href="/rpkp/">RetroPsychoKinesis Project Home</a></h3>
<p>
</body>
</html>
EOH
        ;
    }

    #   Calculate aggregate statistics for a table, append them, and
    #   close out the table.

    sub end_table {
        local($tsd);

        $tsd = (($Ttries / 2) - $Thits) / &standard_deviation($Ttries);
        if ($tsd < 0) {
            $tsd = -$tsd;
        }
        $tsd = sprintf("%.2f", $tsd);
        if ($Thits < ($Ttries / 2)) {
            $tsd = "<font color=\"#FF0000\">" . $tsd . "</font>";
        }
        print <<"EOH"
<tr>
<th colspan=5 align=left>Total
<td align=center>$Thits/$Ttries
<td align=center>$tsd
</tr>
</table>
EOH
        ;
    }

    #   Quote a string, inserting "\" escapes for single
    #   and quotes and "\"

    sub quote_string {
        local($s) = @_;
        $s =~ s/\\/\\\\/;
        $s =~ s/\"/\\\"/;
        $s =~ s/\'/\\\'/;
        '"' . $s . '"';
    }

    #   Count total bits and one bits in a hexadecimal stream

    sub count_bits {
        local($s) = @_;
        local($tbits, $obits, $n);

        $tbits = $obits = 0;
        while (length($s) > 0) {
            $obits += $nbits[hex(substr($s, 0, 1))];
            $tbits += 4;
# print "$obits $tbits " . substr($s, 0, 1) . "\n";
            $s = substr($s, 1);
        }
        ($tbits, $obits);
    }

    #   Compute standard deviation for 0.5 probability binomial
    #   distribution of n attempts.

    sub standard_deviation {
        local($n) = @_;
        sqrt($n / 4);                 # Remember, Perl division result is floating !
    }

    #   Compute combinations of n things taken k at a time

    sub combinations {
        local($n, $k) = @_;
        local($i, $cd, $f);

        $f = 1.0;
        if ($k > ($n - $k)) {
            $cd = $n - $k;
            for ($i = $k + 1; $i <= $n; $i++) {
                if ($i > ($n - $k)) {
                    $f *= $i;
                }
                if ($cd > 1) {
                    $f /= $cd--;
                }
            }
        } else {
            $cd = $k;
            for ($i = ($n - $k) + 1; $i <= $n; $i++) {
                if ($i > $k) {
                    $f *= $i;
                }
                if ($cd > 1) {
                    $f /= $cd--;
                }
            }
        }
        $f;
    }

    sub chance {
        local($hits, $trials, $prob) = @_;

        if ($trials != 1024) {
            return "Trials ($trials) != 1024";
        }
        if ($prob != 0.5) {
            return "Prob ($prob) != 0.5";
        }
        if ($hits > 512) {
            $hits = $trials - $hits;
        }
        $cprob[$hits];
    }

    #   Compute a to the b power

    sub pow {
        local($a, $b) = @_;
 
        exp(log($a) * $b);
    }

    #   Probability of hits successes in trials attempts with a
    #   constant probability of prob on each attempt.

    sub probability {
        local($hits, $trials, $prob) = @_;
 
        &combinations($trials, $hits) * &pow($prob, $hits) *
            &pow(1 - $prob, $trials - $hits);
    }

    #   Cumulative probability of hits successes in trials
    #   attempts with a constant probability of prob on each attempt.

    sub cumulative {
        local($hits, $trials, $prob) = @_;
        local($t, $i);
 
        $t = 0;
        for ($i = 0; $i < $hits; $i++) {
            $t += &probability($i, $trials, $prob);
        }
        $t;
    }
