/*

               by John Walker  --  kelvin@fourmilab.ch
                       http://www.fourmilab.ch/

                This program is in the public domain.

    The commented out stuff is a futile attempt to optimise performance by
    only repainting the modified region.  Java's Graphics package
    clipRect() is so flaky, and so platform-sensitive, that we dare not
    use it at the present time.

*/

import java.awt.*;

public class Cadran extends VisualFeedback {

    protected Color c_dial, c_frame, c_zero, c_pos, c_neg, c_ticks;
    protected int cx, cy, hour_hand, minute_hand, hand_base;
    protected Image face;
    private static final double minuteToRadian = Math.PI / 30;
    private int xPoints[][], yPoints[][];
    private static int px, py;
//  protected Rectangle oldExt = new Rectangle();
//  protected int twiddle = 0;
    protected Dimension window;
    
    // Constructor

    public Cadran(int width, int height) {
        super(Math.min(width, height), Math.min(width, height));
        c_frame = new Color(205, 173, 0);
        c_zero = Color.yellow;
        c_pos = Color.blue;
        c_neg = Color.red;
        c_dial = new Color(220, 220, 220);
        c_ticks = Color.green;
        xPoints = new int[2][4];
        yPoints = new int[2][4];
    }

    /*  PROJECT  --  Given an orientation on the dial expressed
                     in minutes and the radius from the centre
                     of the face, return the Cartesian pixel
                     co-ordinates in px and py.  */

    private final void project(int minutes, int radius) {
        minutes = 75 - minutes;       // Move origin to Y axis
        px = cx + (int) (Math.cos(minutes * minuteToRadian) * radius);
        py = cy - (int) (Math.sin(minutes * minuteToRadian) * radius);
//      oldExt.add(px, py);
    }

    /*  PLOTHAND  --  Plot position of a hand on the dial, given the
                      graphics context, the orientation of the hand in
                      minutes (0 to 60), and the length of the hand in
                      pixels.  The width of the base of the hand is
                      precomputed based on the size of the applet.
                      Plotting must be done before we actually draw
                      the hand thanks to the idiot way clipRect
                      works in Graphics.  */

    protected void plotHand(Graphics g, int which, int minutes, int length) {
        project(minutes, length);
        xPoints[which][0] = px;
        yPoints[which][0] = py;

        project(minutes - 15, hand_base);
        xPoints[which][1] = px;
        yPoints[which][1] = py;

        project(minutes - 30, (hand_base * 3) / 2);
        xPoints[which][2] = px;
        yPoints[which][2] = py;

        project(minutes + 15, hand_base);
        xPoints[which][3] = px;
        yPoints[which][3] = py;
    }

    /*  PAINTHAND  --  Actually paint a hand, previously plotted by
                       plotHand().  */

    protected void paintHand(Graphics g, int which) {
        Color c = g.getColor();

        g.fillPolygon(xPoints[which], yPoints[which], 3);
        g.setColor(c.darker());
        xPoints[which][1] = xPoints[which][3];
        yPoints[which][1] = yPoints[which][3];
        g.fillPolygon(xPoints[which], yPoints[which], 3);
        g.setColor(c);
    }

    //  TICKIMAGE  --  Create image of hour ticks

    protected Image tickImage(int radius) {
        Image i_tick = createImage(radius * 2, radius * 2);
        Graphics b;
        int i, cr, cg, cb;

        b = i_tick.getGraphics();
        b.setColor(c_dial);
        b.fillRect(0, 0, radius * 2, radius * 2);
        cr = c_ticks.getRed();
        cg = c_ticks.getGreen();
        cb = c_ticks.getBlue();
        for (i = radius; i > 0; i--) {
            double shade = Math.cos((Math.PI * 0.42) *
                                    ((double) i) / radius);

            b.setColor(new Color((int) (cr * shade), (int) (cg * shade),
                                 (int) (cb * shade)));
            b.fillOval(radius - i, radius - i, i * 2, i * 2);
        }
        return i_tick;
    }

    //  PAINTWINDOW  --  Paint the component
    
    public void paintWindow(Graphics g) {
        int zamp, hours, minutes;

        zamp = currentScore % 720;
        hours = zamp / 60;
        minutes = zamp % 60;

        if (face == null) {
            Graphics fg;
            int i, dial_width, tick_size, tick_radius;
            Image tick;

            dial_width = Math.max(size.height / 25, 4);
            tick_size = Math.max(size.height / 32, 2);
            tick_radius = size.height / 16;
            hour_hand = size.height / 4;
            minute_hand = (size.height * 6) / 16;
            hand_base = Math.max(size.height / 30, 3);
            cx = size.width / 2;
            cy = size.height / 2;

            //  Create clock face image

            face = createImage(size.width, size.height);
            fg = face.getGraphics();

            //  Window background
            fg.setColor(c_background);
            fg.fillRect(0, 0, size.width, size.height);

            //  Frame at outside of clock
//          fg.setColor(c_frame);
//          fg.fillOval(0, 0, size.width, size.height);
            {
                int cr, cg, cb;

                cr = c_frame.getRed();
                cg = c_frame.getGreen();
                cb = c_frame.getBlue();
                for (i = 0; i < dial_width; i++) {
                    double shade = Math.cos((Math.PI * 0.42) *
                                         ((double) i) / (dial_width / 2));

                    fg.setColor(new Color((int) (cr * shade), (int) (cg * shade),
                                 (int) (cb * shade)));
                    fg.fillOval(i, i, size.width - (2 * i), size.height - (2 * i));
                }

            }

            //  Dial background
            fg.setColor(c_dial);
            fg.fillOval(dial_width / 2, dial_width / 2,
                         size.width - dial_width,
                         size.height - dial_width);

            //  Draw hour ticks around the face.

            tick = tickImage(tick_size / 2);
            for (i = 0; i < 12; i++) {
                project(i * 5, cy - tick_radius);
                fg.drawImage(tick, px - tick_size / 2, py - tick_size / 2, this);
            }

            //  Reset extent box for painting

//          oldExt.reshape(0, 0, window.width, window.height);
//          twiddle = 0;
            fg.dispose();
        }
// g.setColor(Color.magenta);                           // Clip region debug
// g.drawLine(size.width / 2, 0, size.width / 2, 9999); // Clip region debug
// g.drawLine(0, size.width / 2, 9999, size.width / 2); // Clip region debug
//      Graphics ng = g.create();
//      if ((++twiddle) > 4) {
//          ng.clipRect(oldExt.x, oldExt.y, oldExt.width + 10, oldExt.height + 10);
//      }
//      ng.drawImage(face, 0, 0, this);
//      ng.dispose();
//      oldExt.reshape(window.width / 2, window.height / 2, 0, 0);
/**/    g.drawImage(face, 0, 0, this);
        g.setColor(currentScore == 0 ? c_zero :
            (currentScore > 0 ? c_pos : c_neg));
        plotHand(g, 0, hours * 5 + (int) (minutes / 12), hour_hand);
        plotHand(g, 1, minutes, minute_hand);
        paintHand(g, 0);
        paintHand(g, 1);
    }

    /*  VALIDATE  --  We need to know when the layout manager
                      changes our size so we can regenerate the
                      face bitmap on the next paint request.  */

    public void validate() {
        int dim;

        super.validate();
        dim = Math.min(size().width, size().height);
        size = new Dimension(dim, dim);
        window = size;
        face = null;
    }
}
