/*

               by John Walker  --  kelvin@fourmilab.ch
                       http://www.fourmilab.ch/

                This program is in the public domain.

*/

import java.applet.*;
import java.awt.*;
import java.util.*;
import runBitSet;

public class Scoring {
    int ntrials = 0, score = 0;
    double sigma = 0, significance = 0;
    boolean hit;

    /*  Constructor  --  Compute score based on the contents of a
                         runBitSet.  */

    Scoring(runBitSet rb, boolean goal)
    {
        ntrials = rb.size();
        score = (goal ? rb.numOne() : rb.numZero()) - rb.mean();
        hit = (score >= 0) ? goal : !goal;
        sigma = rb.resultStandardDeviation();
        significance = rb.resultSignificance();
    }

    //  Centre text in a window of a given size

    static final void centreText(String s, Graphics g, int width, int y)
    {
        g.drawString(s, (width - g.getFontMetrics().stringWidth(s)) / 2, y);
    }

    /*  PAINTQUIT  --  Paint message confirming that user has quit.  */

    public void paintQuit(Graphics g, int width, int height)
    {
        Font f = new Font("TimesRoman", Font.PLAIN, 18);
        FontMetrics fm;
        int h = 10, l;

        g.setFont(f);
        fm = g.getFontMetrics();
        l = (height - fm.getHeight()) / 2;

        g.setColor(Color.lightGray);
        g.fillRect(0, 0, width, height);

        g.setColor(Color.red);
        centreText("Quit at user request.", g, width, l);
    }

    /*  PAINTSCORE  --  Paint the scoreboard into a given graphics
                        context.  */

    public void paintScore(Graphics g, int width, int height)
    {
        Font f = new Font("TimesRoman", Font.PLAIN, 18);
        FontMetrics fm;
        int h = 10, l;

        g.setFont(f);
        fm = g.getFontMetrics();
        l = fm.getHeight();

        g.setColor(Color.lightGray);
        g.fillRect(0, 0, width, height);

        g.setColor(Color.black);
        centreText("Results", g, width, l);

        g.drawString("Final score: " + score, h, l * 3);
        g.drawString(Double.toString(sigma) + " standard deviations", h, l * 4);
        g.drawString("Significance: " + Double.toString(significance) + ".  A score", h, l * 5);
        g.drawString("of this many " + (hit ? "hit" : "misse") + "s can be", h, l * 6);
        g.drawString("expected by chance every " + Integer.toString((int) (1 / significance)), h, l * 7);
        g.drawString("runs.", h, l * 8);
    }
}
