/*

               by John Walker  --  kelvin@fourmilab.ch
                       http://www.fourmilab.ch/

                This program is in the public domain.

*/

import java.applet.*;
import java.awt.*;
import java.util.*;
import java.net.*;
import VisualFeedback;
import runBitSet;

public class experiment extends Applet implements Runnable {
    static String appName = "experiment";       // Experiment name
    static String Version = "Version 1.0";
    final static int Mode_demo = 0,             // Operating mode
                     Mode_practice = 1,
                     Mode_record = 2,
                     Mode_constant = 3;

    VisualFeedback display;
    AudioClip sound;
    Thread animator;
    barMeter howFar;
    int progressBar = -1;
    int barLeft, barRight;
    String audioFile = "bonk.au";
    Label slabel = null, slabelt = null, vlabel;
    int rate = 100; 
    int score = 0;
    int supd = 500 / rate, sn = 0;
    Rectangle r;
    Button start_stop;
    boolean error = false, running = true, quitIt = false;
    int nrand = 1024, inrand = -1;    // Random bits per run
    int howOften = 0, thatOften = 0;  /* If thatOften is nonzero (having been
                                         set so by the constructor of a
                                         subclass) bits are returned from
                                         the random number generator only
                                         every thatOften passes through the
                                         animation loop.  This allows
                                         animations which consume bits at
                                         any desired rate while continuing
                                         to run. */
    Random rng;
    int randomOffset = 0;
    runBitSet rb = null;
    Scoring finalScore;

    //  Parameters from applet invocation

    int mode;
    boolean ones, silent, show_version, show_score;
    String Email;

    /*  RANDOMSOURCE  --  Return next bit from random number source.
                          Returns 0 or 1 for next bit, -1 at end of
                          file.  */

    protected int randomsource() {
        if (progressBar != -1) {
            howFar.moveBar(progressBar, barLeft, barRight -
                                (((barRight - barLeft) * nrand) / inrand));
        }
        if (mode != Mode_constant && (--nrand < 0)) {
            return -1;
        }
        if (mode == Mode_constant) {
            return (rng.nextInt() & 0x20) == 0 ? 0 : 1;
        }

        return rb.get(randomOffset++) ? 1 : 0;
    }

    //  RUN  --  Main thread processing

    public void run() {
        while (true) {

            if (running) {
                if (thatOften == 0 || (((++howOften) % thatOften) == 0)) {
                    int gloox = randomsource();

                    //  Update the score and notify the display component

                    if (mode != Mode_constant && (gloox == -1 || quitIt)) {
                        start_stop.hide();
                        if (howFar != null) {
                            howFar.hide();
                        }
                        if (slabel != null) {
                            slabelt.hide();
                            slabel.hide();
                        }
                        display.postScore(finalScore, quitIt);
                        running = false;
                     } else {
                        if (gloox == 0) {
                          score--;
                        } else  {
                          score++;
                        }
                        if (show_score && (sn++ > supd)) {
                              slabel.setText("" + ((ones ? 1 : -1) * score));
                            sn = 0;
                        }
                    }
                }
                display.setRunningScore(score);
            }

            try {
                animator.sleep(rate);
            } catch (InterruptedException e) {
                return;
            }
        }
    }

    /*  GETPARAMETERINFO  --  Standard applet method that returns an
                              array of arrays of strings describing the
                              parameters this applet accepts.  We append
                              any parameters accepted by the visual
                              feedback program to the end of the list.
                              In doing so we make the assumption that
                              init() will always be called before
                              getParameterInfo().  */

    public String[][] getParameterInfo() {
        String[][] info = {
            { "Goal", "1:Ones/0:Zeroes", "Goal of experiment, default 1:Ones" },
            { "Email", "String", "Subject's E-mail address, if supplied" },
            { "Sound", "-s/-n", "Generate sound feedback?  Default -n" },
            { "Mode", "Demo/Practice/Record/Constant", "Record results mode, default Demo" },
            { "Version", "Off/On", "Show version in header?  Default Off" },
            { "Score", "On/Off", "Show score during run?  Default On" },
            { "RandomData", "Hex string", "Random data for Record and Practice runs" }
        };
        String[][] merged = new String[info.length + display.parameters.length][3];
        int i;

        System.arraycopy(info, 0, merged, 0, info.length);
        if (display.parameters.length > 0) {
            System.arraycopy(display.parameters, 0, merged, info.length, display.parameters.length);
        }
        return merged;
    }

    /*  GETAPPLETINFO  --  Return information about this applet.  */

    public String getAppletInfo() {
        return "RPKP " + appName +
               " experiment by John Walker.  Public domain.";
    }

    /*  PROCESS_PARAMETERS  --  Process applet parameters.  All parameters
                                are kept in instance variables of the
                                applet.  */

    protected void process_parameters() {
        String s;

        //  Goal of experiment

        s = getParameter("Goal");
        if (s == null) {
            s = "1";
        }
        ones = s.startsWith("1");

        //  Subject's E-mail address

        Email = getParameter("Email");
        if (Email == null) {
            Email = "";
        }

        //  Generate sound ?

        s = getParameter("Sound");
        if (s == null) {
            s = "Silent";
        }
        silent = !s.equalsIgnoreCase("-s");

        //  Experiment mode

        s = getParameter("Mode");
        if (s == null) {
            s = "Demo";
        }
        if (s.equalsIgnoreCase("Demo")) {
            mode = Mode_demo;
        } else if (s.equalsIgnoreCase("Practice")) {
            mode = Mode_practice;
        } else if (s.equalsIgnoreCase("Record")) {
            mode = Mode_record;
        } else if (s.equalsIgnoreCase("Constant")) {
            mode = Mode_constant;
        }

        //  Show version of program ?

        s = getParameter("Version");
        if (s == null) {
            s = "Off";
        }
        show_version = s.equalsIgnoreCase("On");

        //  Show score during run ?

        s = getParameter("Score");
        if (s == null) {
            s = "On";
        }
        show_score = s.equalsIgnoreCase("On");

        /*  Random data stream.  This is supplied in hexadecimal.
            Parsing is delegated to the runBitSet object created
            from the data stream, which is well known to ignore
            white space in its food dish.  */

        s = getParameter("RandomData");
        if (s != null) {
            rb = new runBitSet(s);
            nrand = rb.size();
        } else {
            s = getParameter("Nrand");
            if (s != null) {
                Integer i;

                try {
                    i = Integer.valueOf(s);
                    if (i.intValue() > 0) {
                        nrand = i.intValue() * 8;
                    }
                } catch (NumberFormatException e) {
                }
            }
        }
    }

    //  MAKEDISPLAY  --  Create visual feedback component

    protected void makeDisplay(int width, int height) {
        display = new VisualFeedback(width, height);
    }

    //  INIT  --  Applet is loaded--create our objects and display them

    public void init() {
        int pheight, i;
        String s;
        GridBagLayout gribble = new GridBagLayout();
        GridBagConstraints gc = new GridBagConstraints();

        setLayout(gribble);
        process_parameters();         // Parse parameters and set mode variables

        //  If demo mode, create pseudorandom bit set

        if (mode == Mode_demo && rb == null) {
            rb = new runBitSet(nrand);
        }

        //  For constant mode, create inexhaustible pseudorandom source

        if (mode == Mode_constant) {
            rng = new Random();
        } else {
            /*  Why compute the score here?  Because doing so (in this
                pre .jar era) forces the Scoring class to be downloaded
                now rather than at the end of the experiment where the
                user will be impatient to see the results and may abort
                the download before they appear.  */
            finalScore = new Scoring(rb, ones);
        }

        if ((mode == Mode_practice || mode == Mode_record) &&
            (rb == null || nrand <= 0)) {
            bomb("random data missing");
        }

        if (error) {
            return;
        }

        r = bounds();
        pheight = r.height;

        makeDisplay(r.width, r.height);

        display.setBackground(getBackground());

        /*  Now allow the visual feedback object to process any
            parameters of its own.  */

        for (i = 0; i < display.parameterCount(); i++) {
            display.setParameter(i, getParameter(display.parameterName(i)));
        }

        if (mode != Mode_constant) {
            start_stop = new Button("Start");
            running = false;
        }

        if (!silent) {
            sound = getAudioClip(getCodeBase(), audioFile);
            /* What's this, you ask?  Well, you see, getAudioClip
               doesn't actually download the clip until you use it
               the first time.  If we don't try to play it here,
               it would cause the display to pause at the time the
               limit was hit for the first time, disrupting the
               otherwise smooth animation.  So, we force it to be
               downloaded here by playing it, then immediately
               stopping playback.  Unfortunately, the stop() method
               does not work on all platforms, resulting in an
               unwanted initial sound. */
            sound.play();
            sound.stop();
            display.setAudioClip(sound);
        }

        if (show_version) {
            vlabel = new Label(Version, Label.CENTER);
        }
        if (show_score) {
            slabelt = new Label("Score: ", Label.RIGHT);
            slabel = new Label("0      ", Label.LEFT);
        }

        /*  All user interface components have now been built;
            lay them out within the applet window.  */

        if (vlabel != null) {
            gc.anchor = GridBagConstraints.CENTER;
            gc.fill = GridBagConstraints.HORIZONTAL;
            gc.gridwidth = GridBagConstraints.REMAINDER;
            gc.weightx = 0.0;
            gribble.setConstraints(vlabel, gc);
            add(vlabel);
        }

        gc.anchor = GridBagConstraints.CENTER;
        gc.fill = GridBagConstraints.BOTH;
        gc.gridwidth = GridBagConstraints.REMAINDER;
        gc.weightx = 0.0;
        gc.weighty = 1;
        gribble.setConstraints(display, gc);
        add(display);
        gc.weighty = 0;

        if (mode != Mode_constant) {
            howFar = new barMeter(r.width, 16, barMeter.HORIZONTAL);
            howFar.setBorder(5, 1);
            howFar.addBar(getBackground(), barMeter.percent(0),
                          barMeter.percent(100), barMeter.percent(100),
                          barMeter.LOWERED, barMeter.VISIBLE);
            barLeft = howFar.pixelsX(2);
            barRight = barMeter.percent(100) - barLeft;
            progressBar = howFar.addBar(Color.cyan, barLeft, barRight,
                                        barMeter.percent(100) - howFar.pixelsY(2),
                                        barMeter.FLAT, barMeter.VISIBLE);
            progressBar = howFar.addBar(Color.blue, barLeft, barLeft,
                                        barMeter.percent(100) - howFar.pixelsY(2),
                                        barMeter.FLAT, barMeter.HIDDEN);
            gc.anchor = GridBagConstraints.CENTER;
            gc.fill = GridBagConstraints.HORIZONTAL;
            gc.gridwidth = GridBagConstraints.REMAINDER;
            gc.weightx = 1.0;
            gribble.setConstraints(howFar, gc);
            add(howFar);
        }

        gc.anchor = GridBagConstraints.CENTER;
        gc.fill = GridBagConstraints.NONE;
        gc.weightx = 1.0;

        if (slabel != null) {
            gc.anchor = GridBagConstraints.EAST;
            gc.gridwidth = GridBagConstraints.RELATIVE;
            gribble.setConstraints(slabelt, gc);
            add(slabelt);
            gc.anchor = GridBagConstraints.WEST;
            gc.gridwidth = GridBagConstraints.REMAINDER;
            gribble.setConstraints(slabel, gc);
            add(slabel);
        }

        if (mode != Mode_constant) {
            gc.anchor = GridBagConstraints.CENTER;
            gc.gridwidth = GridBagConstraints.REMAINDER;
            gribble.setConstraints(start_stop, gc);
            add(start_stop);
        }

        animator = new Thread(this);
    }

    //  ACTION  --  Respond to events from child components

    public boolean action(Event e, Object arg) {
        if (e.target == start_stop) {
            if (mode == Mode_record) {
                /* A record mode session cannot be interrupted,
                   so we disable the Start/Stop button. */
                start_stop.disable();
            } else {
                if (running) {
                    quitIt = true;
                    return true;
                }

                /* Demo or Practice sessions can be interrupted
                   early by pressing the Start/Quit button.  Change
                   the legend on the button to reflect its next
                   action. */

                start_stop.setLabel("Quit");
            }
            running = !running;
            if (running && progressBar != -1) {
                howFar.showBar(progressBar, barMeter.VISIBLE);
                if (inrand == -1) {
                    inrand = nrand;
                }
            }
            return true;
        }
        return false;
    }

    //  START  --  Applet loading complete; start animation thread

    public void start() {
        if (animator != null) {
            if (animator.isAlive()) {
                animator.resume();
            } else {
                animator.start();
            }
        }
    }

    //  STOP  --  We're leaving this page--stop animation thread

    public void stop() {
        if (animator != null) {
            animator.suspend();
        }
    }

    //  DESTROY  --  Endsville--destroy the animation thread

    public void destroy() {
        if (animator != null) {
            animator.stop();
        }
    }

    //  BOMB  --  Show error message and abort

    private void bomb(String err) {
        add(new Label(appName + " applet error:\n"));
        add(new Label(err));
        showStatus(appName + " applet error: " + err);
        error = true;
    }
}
