/*

               by John Walker  --  kelvin@fourmilab.ch
                       http://www.fourmilab.ch/

                This program is in the public domain.

*/

import java.util.*;
import java.awt.*;

public class pipeorgan extends experiment {
    protected Organ bdisp;
    Button pause_run, refresh = null;
    TextField nBits;
    int nruns = 0;
    boolean paused = false;
    Panel pLabel;
    static final String pauseLegend = " Pause ",
                        resumeLegend = "Resume",
                        startLegend = "Start",
                        stopLegend = " Stop ",
                        refreshLegend = "Repaint";

    //  INIT  --  Applet is loaded--create our objects and display them

    public void init() {
        int pheight, i;
        String s;
        GridBagLayout gribble = new GridBagLayout();
        GridBagConstraints gc = new GridBagConstraints();
        Panel flipPanel;

        appName = "pipeorgan";        // Experiment name
        setLayout(gribble);
        process_parameters();         // Parse parameters and set mode variables
        nBits = new TextField("32", 4);
        flipPanel = new Panel();
        flipPanel.add(new Label("Flips/run:"));
        flipPanel.add(nBits);

        rng = new Random();

        r = bounds();
        pheight = r.height;

        makeDisplay(r.width, r.height);

        display.setBackground(getBackground());

        /*  Now allow the visual feedback object to process any
            parameters of its own.  */

        for (i = 0; i < display.parameterCount(); i++) {
            display.setParameter(i, getParameter(display.parameterName(i)));
        }

        if (mode != Mode_constant) {
            start_stop = new Button(startLegend);
            pause_run = new Button(pauseLegend);
            running = false;
        }

        if (show_score) {
            slabelt = new Label("Runs: ", Label.RIGHT);
            slabel = new Label("0", Label.LEFT);
        }

        /*  All user interface components have now been built;
            lay them out within the applet window.  */

        gc.anchor = GridBagConstraints.CENTER;
        gc.fill = GridBagConstraints.BOTH;
        gc.gridwidth = GridBagConstraints.REMAINDER;
        gc.weightx = 0.0;
        gc.weighty = 1;
        gribble.setConstraints(display, gc);
        add(display);
        gc.weighty = 0;

        gc.anchor = GridBagConstraints.CENTER;
        gc.fill = GridBagConstraints.NONE;
        gc.weightx = 1.0;

        if (slabel != null) {
            pLabel = new Panel();

            pLabel.add(slabelt);
            pLabel.add(slabel);
            gc.anchor = GridBagConstraints.CENTER;
            gc.gridwidth = GridBagConstraints.REMAINDER;
            gribble.setConstraints(pLabel, gc);
            add(pLabel);
/*
            gc.gridwidth = GridBagConstraints.CENTER;
            gribble.setConstraints(slabelt, gc);
            add(slabelt);
            gc.anchor = GridBagConstraints.CENTER;
            gc.gridwidth = GridBagConstraints.REMAINDER;
            gribble.setConstraints(slabel, gc);
            add(slabel);
*/
        }

        if (mode != Mode_constant) {
            gc.anchor = GridBagConstraints.CENTER;
            gc.gridwidth = 1;
            gribble.setConstraints(start_stop, gc);
            add(start_stop);
            gribble.setConstraints(pause_run, gc);
            add(pause_run);
            add(flipPanel);
            pause_run.disable();
            gc.gridwidth = GridBagConstraints.REMAINDER;
            refresh = new Button(refreshLegend);
            gribble.setConstraints(refresh, gc);
            add(refresh);
        }

        supd = 10;
        bdisp.setView();

        animator = new Thread(this);
    }

    protected void makeDisplay(int width, int height) {
        display = (VisualFeedback) (bdisp = new Organ(width, height));
    }


    //  RUN  --  Main thread processing

    public void run() {
        while (true) {
            if (quitIt) {
                running = false;
                quitIt = false;
            }
            if (running && !paused) {
                int n;

                score = 0;
                for (n = 0; n < bdisp.bitsPerExp; n++) {

                    //  Update the score and notify the display component

                    int gloox = randomsource();

                     if (gloox == 1) {
                        score++;
                    }
                }
                nruns++;
                if (running && show_score && ((nruns <= 10) || (++sn >= supd))) {
                    String s = "" + nruns;
                    boolean longer = s.length() > slabel.getText().length();

                    slabel.setText(s);
                    if (longer) {
                        invalidate();
                              
                        pLabel.getParent().validate();
                    }
                    sn = 0;
                }
                display.setRunningScore(score);
            }

            try {
                animator.sleep(rate);
            } catch (InterruptedException e) {
                return;
            }
        }
    }

    //  SETBITS  --  Set number of bits from bdisp component value

    private void setBits() {
        int b;

        try {
            b = Integer.parseInt(nBits.getText());
            if (b < 4) {
                b = 4;
                nBits.setText("4");
            } else if (b > 1024) {
                b = 1024;
                nBits.setText("1024");
            }
        } catch (NumberFormatException ex) {
            b = 32;
            nBits.setText("32");
        }
        bdisp.reset(b);
    }

    //  ACTION  --  Respond to events from child components

    public boolean action(Event e, Object arg) {
        if (e.target == start_stop) {
            if (running) {
                quitIt = true;
                pause_run.disable();
            } else {
                setBits();
                nruns = 0;
                sn = 0;
                slabel.setText("0");
                pause_run.setLabel(pauseLegend);
                pause_run.enable();
                paused = false;
            }
            start_stop.setLabel(running ? startLegend : stopLegend);
            running = !running;
            nBits.setEditable(!running);
            return true;
        } else if (e.target == pause_run) {
            if (running) {
                paused = !paused;
                pause_run.setLabel(paused ? resumeLegend : pauseLegend);
            }
        } else if (e.target == nBits) {
            if (e.id == Event.ACTION_EVENT) {
                int b;

                setBits();
                bdisp.repaint();
            }
        } else if (refresh != null && e.target == refresh) {
            bdisp.repair();
        }
        return false;
    }

    /*  RANDOMSOURCE  --  Override to as to use locally-generated
                          bits in all modes.  */

    protected int randomsource() {
        return (rng.nextInt() & 0x20) == 0 ? 0 : 1;
    }
}
